// Copyright 2021 Chaos Mesh Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package genericwebhook

import (
	"reflect"
	"strings"

	"k8s.io/apimachinery/pkg/util/validation/field"
)

type Defaulter interface {
	Default(root interface{}, field *reflect.StructField)
}

// Default would walk through all the fields of target struct recursively, and set the default value which declared with struct tag "default".
//
// Parameter obj should be a pointer to a data struct.
//
// Default should return an empty field.ErrorList.
func Default(obj interface{}) field.ErrorList {
	// TODO: how to resolve invalid input, for example: obj is a pointer to pointer
	errorList := field.ErrorList{}

	root := obj
	walker := NewFieldWalker(obj, func(path *field.Path, obj interface{}, field *reflect.StructField) bool {
		webhookAttr := ""
		if field != nil {
			webhookAttr = field.Tag.Get("webhook")
		}
		attributes := strings.Split(webhookAttr, ",")

		webhook := ""
		nilable := false
		if len(attributes) > 0 {
			webhook = attributes[0]
		}
		if len(attributes) > 1 {
			nilable = attributes[1] == "nilable"
		}

		defaulter := getDefaulter(obj, webhook, nilable)
		if defaulter != nil {
			defaulter.Default(root, field)
		}

		return true
	})
	walker.Walk()

	return errorList
}

func getDefaulter(obj interface{}, webhook string, nilable bool) Defaulter {
	// There are two possible situations:
	// 1. The field is a value (int, string, normal struct, etc), and the obj is the reference of it.
	// 2. The field is a pointer to a value or a slice, then the obj is itself.

	val := reflect.ValueOf(obj)

	if defaulter, ok := obj.(Defaulter); ok {
		if nilable || !val.IsZero() {
			return defaulter
		}
	}

	if webhook != "" {
		webhookImpl := webhooks[webhook]

		v := val.Convert(webhookImpl).Interface()
		if defaulter, ok := v.(Defaulter); ok {
			if nilable || !val.IsZero() {
				return defaulter
			}
		}
	}

	return nil
}
